@@ -1,9 +1,9 @@
 from base64 import b64encode
 from glob import glob
 from ipaddress import ip_network
-from kvirt.common import success, error, pprint, info2, container_mode, warning, fix_typos
-from kvirt.common import get_oc, pwd_path, get_installer_rhcos, get_ssh_pub_key, start_baremetal_hosts, olm_app
-from kvirt.common import deploy_cloud_storage
+from kvirt.common import success, error, pprint, info2, container_mode, warning, fix_typos, olm_app
+from kvirt.common import get_oc, pwd_path, get_installer_rhcos, get_ssh_pub_key, start_baremetal_hosts_with_iso
+from kvirt.common import deploy_cloud_storage, patch_ingress_controller_wildcard
 from kvirt.cluster.openshift import get_ci_installer, get_downstream_installer, get_installer_version
 from kvirt.cluster.openshift import same_release_images, process_apps, offline_image
 import json
@@ -20,8 +20,8 @@ from urllib.parse import urlparse
 from urllib.request import urlopen, Request
 from yaml import safe_dump, safe_load
 
-virt_providers = ['kvm', 'kubevirt', 'ovirt', 'openstack', 'vsphere', 'proxmox']
-cloud_providers = ['aws', 'azure', 'gcp', 'ibm']
+virt_providers = ['kvm', 'kubevirt', 'ovirt', 'openstack', 'proxmox', 'vsphere']
+cloud_providers = ['aws', 'azure', 'gcp', 'ibm', 'hcloud']
 
 
 def update_hypershift_etc_hosts(cluster, domain, ingress_ip):
@@ -68,7 +68,7 @@ def get_info(url, user, password):
         model = 'virtual'
         user = user or 'fake'
         password = password or 'fake'
-        url = url.replace('http', 'redfish-virtualmedia+http')
+        url = url.replace('http', 'redfish-virtualmedia')
         return url, user, password
     oem_url = f"https://{url}" if '://' not in url else url
     p = urlparse(oem_url)
@@ -171,14 +171,55 @@ def handle_baremetal_iso(config, plandir, cluster, data, baremetal_hosts=[]):
         return f'http://{svcip}:{svcport}/{cluster}-worker.iso'
 
 
+def process_nodepools(config, plandir, namespace, cluster, platform, nodepools, image, overrides, manifests=[]):
+    clusterdir = os.path.expanduser(f"~/.kcli/clusters/{cluster}")
+    nodepools = [{'name': entry} if isinstance(entry, str) else entry for entry in nodepools]
+    nodepool_data = {'nodepools': nodepools, 'namespace': namespace, 'cluster': cluster,
+                     'nodepool_image': image, 'platform': platform, 'workers': overrides.get('workers', 2)}
+    kubevirt_data = {'numcpus': overrides.get('numcpus', 8), 'memory': overrides.get('memory', 6144),
+                     'disk_size': overrides.get('disk_size', 30)}
+    nodepool_data.update(kubevirt_data)
+    if manifests:
+        nodepool_data['manifests'] = manifests
+    nodepoolfile = config.process_inputfile(cluster, f"{plandir}/nodepools.yaml", overrides=nodepool_data)
+    with open(f"{clusterdir}/nodepools.yaml", 'w') as f:
+        f.write(nodepoolfile)
+    cmcmd = f"oc create -f {clusterdir}/nodepools.yaml"
+    call(cmcmd, shell=True)
+    if platform is not None:
+        return
+    for entry in nodepools:
+        nodepool = entry['name']
+        pprint(f"Waiting before ignition data for nodepool {nodepool} is available")
+        user_data = f"user-data-{cluster}"
+        cmd = f"until oc -n {namespace}-{cluster} get secret | grep {user_data} >/dev/null 2>&1 ; do sleep 1 ; done"
+        call(cmd, shell=True)
+        ignition_data = {'namespace': namespace, 'cluster': cluster, 'nodepool': nodepool}
+        ignitionscript = config.process_inputfile(cluster, f"{plandir}/ignition.sh", overrides=ignition_data)
+        with open(f"{clusterdir}/ignition_{nodepool}.sh", 'w') as f:
+            f.write(ignitionscript)
+        ignition_worker = f"{clusterdir}/nodepool_{nodepool}.ign"
+        timeout = 0
+        while True:
+            if os.path.exists(ignition_worker):
+                break
+            else:
+                sleep(30)
+                timeout += 30
+                if timeout > 300:
+                    msg = "Timeout trying to retrieve worker ignition"
+                    return {'result': 'failure', 'reason': msg}
+            call(f'bash {clusterdir}/ignition_{nodepool}.sh', shell=True)
+
+
 def scale(config, plandir, cluster, overrides):
     storedparameters = overrides.get('storedparameters', True)
     provider = config.type
     plan = cluster
-    data = {'cluster': cluster, 'kube': cluster, 'kubetype': 'hypershift', 'namespace': 'clusters'}
+    data = {'cluster': cluster, 'kube': cluster, 'kubetype': 'hypershift', 'namespace': 'clusters',
+            'disk_size': 30, 'numcpus': 8, 'memory': 16384}
     data['basedir'] = '/workdir' if container_mode() else '.'
     cluster = data['cluster']
-    namespace = data['namespace']
     clusterdir = os.path.expanduser(f"~/.kcli/clusters/{cluster}")
     if not os.path.exists(clusterdir):
         warning(f"Creating {clusterdir} from your input (auth creds will be missing)")
@@ -186,35 +227,50 @@ def scale(config, plandir, cluster, overrides):
         overrides['cluster'] = cluster
         overrides['clusterdir'] = clusterdir
         plan = overrides.get('plan') or plan
-        if 'ingress_ip' not in overrides and provider != 'kubevirt':
-            msg = "Missing ingress_ip..."
-            return {'result': 'failure', 'reason': msg}
-        domain = overrides.get('domain')
-        if domain is None:
-            msg = "Missing domain..."
-            return {'result': 'failure', 'reason': msg}
-        if 'management_ingress_domain' not in overrides:
-            overrides['management_ingress_domain'] = f'apps.{cluster}.{domain}'
-        os.mkdir(clusterdir)
-        ignitionscript = config.process_inputfile(cluster, f"{plandir}/ignition.sh", overrides=overrides)
-        with open(f"{clusterdir}/ignition.sh", 'w') as f:
-            f.write(ignitionscript)
-        call(f'bash {clusterdir}/ignition.sh', shell=True)
-    old_assisted_vms_number = 2
+        namespace = data['namespace']
+        assisted = cluster in os.popen(f'oc get -n {namespace}-{cluster} infraenv {cluster}').read().strip()
+        if not assisted:
+            if 'ingress_ip' not in overrides and provider != 'kubevirt':
+                msg = "Missing ingress_ip..."
+                return {'result': 'failure', 'reason': msg}
+            domain = overrides.get('domain')
+            if domain is None:
+                msg = "Missing domain..."
+                return {'result': 'failure', 'reason': msg}
+            if 'management_ingress_domain' not in overrides:
+                overrides['management_ingress_domain'] = f'apps.{cluster}.{domain}'
+            ignitionscript = config.process_inputfile(cluster, f"{plandir}/ignition.sh", overrides=overrides)
+            with open(f"{clusterdir}/ignition.sh", 'w') as f:
+                f.write(ignitionscript)
+            call(f'bash {clusterdir}/ignition.sh', shell=True)
+        else:
+            os.mkdir(clusterdir)
+            data['assisted'] = True
+    old_extra_nodepools = []
     if storedparameters and os.path.exists(f"{clusterdir}/kcli_parameters.yml"):
         with open(f"{clusterdir}/kcli_parameters.yml", 'r') as install:
             installparam = safe_load(install)
             data.update(installparam)
             plan = installparam.get('plan', plan)
-            old_assisted_vms_number = installparam.get('assisted_vms_number', old_assisted_vms_number)
+            old_extra_nodepools = installparam.get('extra_nodepools', old_extra_nodepools)
     data.update(overrides)
+    namespace = data['namespace']
+    nodepools = [p for p in data.get('extra_nodepools') if p not in old_extra_nodepools]
+    if nodepools:
+        platform = data.get('platform')
+        nodepool_cmd = "oc get nodepool -n %s %s -o jsonpath='{.spec.release.image}'" % (namespace, cluster)
+        nodepool_image = os.popen(nodepool_cmd).read()
+        process_nodepools(config, plandir, namespace, cluster, platform, nodepools, nodepool_image, overrides)
+    nodepool = overrides.get('nodepool') or cluster
+    if os.popen(f"oc get nodepool -n {namespace} {nodepool} -o name").read() == "":
+        msg = f"Issue with nodepool {nodepool}..."
+        return {'result': 'failure', 'reason': msg}
     if os.path.exists(f"{clusterdir}/kubeconfig.mgmt"):
         os.environ['KUBECONFIG'] = f"{clusterdir}/kubeconfig.mgmt"
     if 'KUBECONFIG' not in os.environ:
         warning("KUBECONFIG not set...Using .kube/config instead")
     platform = data.get('platform')
     assisted = platform == 'assisted'
-    assisted_vms = data.get('assisted_vms')
     kubevirt = platform == 'kubevirt'
     with open(f"{clusterdir}/kcli_parameters.yml", 'w') as paramfile:
         safe_dump(data, paramfile, default_flow_style=False, encoding='utf-8', allow_unicode=True)
@@ -222,42 +278,42 @@ def scale(config, plandir, cluster, overrides):
     worker_overrides = data.copy()
     workers = worker_overrides.get('workers', 2)
     if kubevirt:
-        cmcmd = f"oc -n {namespace}-{cluster} scale nodepool {cluster} --replicas {workers}"
+        cmcmd = f"oc -n {namespace}-{cluster} scale nodepool {nodepool} --replicas {workers}"
         call(cmcmd, shell=True)
         return {'result': 'success'}
     os.chdir(os.path.expanduser("~/.kcli"))
     old_baremetal_hosts = installparam.get('baremetal_hosts', [])
     new_baremetal_hosts = overrides.get('baremetal_hosts', [])
-    assisted_vms_number = overrides.get('workers') or data['assisted_vms_number']
-    if assisted_vms and assisted_vms_number != old_assisted_vms_number:
-        if assisted_vms_number < old_assisted_vms_number:
-            msg = "New assisted_vms_number should be greater than old one"
-            return {'result': 'failure', 'reason': msg}
-        if data.get('disk_size', 30) < 200:
-            data['disk_size'] = 200
-        if data.get('numcpus', 8) < 16:
-            data['numcpus'] = 16
-        if data.get('memory', 16384) < 20480:
-            data['memory'] = 20480
-        worker_threaded = data.get('threaded', False) or data.get('assisted_vms_threaded', False)
-        result = config.plan(plan, inputfile=f'{plandir}/kcli_plan_assisted.yml', overrides=data,
-                             threaded=worker_threaded)
-        if result['result'] != 'success':
-            return result
-        vms = result['newvms']
+    if assisted:
         ksushy_ip = data['assisted_vms_ksushy_ip']
-        ksushy_url = f'http://{ksushy_ip}:9000/redfish/v1/Systems/{config.client}'
-        virtual_hosts = []
-        for vm in vms:
-            new_mac = config.k.info(vm)['nets'][0]['mac']
-            virtual_hosts.append({'url': f'{ksushy_url}/{vm}', 'mac': new_mac})
-        new_baremetal_hosts.extend(virtual_hosts)
+        ksushy_url = f'https://{ksushy_ip}:9000/redfish/v1/Systems/{config.client}'
+        old_physical_baremetal_hosts = [h for h in old_baremetal_hosts if ksushy_ip not in h['url']]
+        assisted_vms_number = workers - len(old_physical_baremetal_hosts + new_baremetal_hosts)
+        if assisted_vms_number > 0:
+            data['assisted_vms_number'] = assisted_vms_number
+            data['disk_size'] = max(data['disk_size'], 200)
+            data['numcpus'] = max(data['numcpus'], 16)
+            data['memory'] = max(data['memory'], 20480)
+            worker_threaded = data.get('threaded', False) or data.get('assisted_vms_threaded', False)
+            result = config.plan(plan, inputfile=f'{plandir}/kcli_plan_assisted.yml', overrides=data,
+                                 threaded=worker_threaded)
+            if result['result'] != 'success':
+                return result
+            vms = result['newvms']
+            virtual_hosts = []
+            for vm in vms:
+                new_mac = config.k.info(vm)['nets'][0]['mac']
+                virtual_hosts.append({'url': f'{ksushy_url}/{vm}', 'mac': new_mac})
+            new_baremetal_hosts.extend(virtual_hosts)
     baremetal_hosts = [entry for entry in new_baremetal_hosts if entry not in old_baremetal_hosts]
     if baremetal_hosts:
         if assisted:
             all_baremetal_hosts = old_baremetal_hosts + baremetal_hosts
             create_bmh_objects(config, plandir, cluster, namespace, all_baremetal_hosts, overrides)
-            cmcmd = f"oc -n {data['namespace']} scale nodepool {cluster} --replicas {len(all_baremetal_hosts)}"
+            replicas_cmd = "oc get nodepool -n %s %s -o jsonpath='{.spec.replicas}'" % (namespace, nodepool)
+            replicas = int(os.popen(replicas_cmd).read()) + len(baremetal_hosts)
+            pprint(f"Setting replicas to {replicas} in nodepool {nodepool}")
+            cmcmd = f"oc -n {data['namespace']} scale nodepool {nodepool} --replicas {replicas}"
             call(cmcmd, shell=True)
             return {'result': 'success'}
         else:
@@ -269,7 +325,9 @@ def scale(config, plandir, cluster, overrides):
                 svcport_cmd = 'oc get svc -n default httpd-kcli-svc -o yaml'
                 svcport = safe_load(os.popen(svcport_cmd).read())['spec']['ports'][0]['nodePort']
                 iso_url = f'http://{svcip}:{svcport}/{cluster}-worker.iso'
-            start_baremetal_hosts(baremetal_hosts, iso_url, overrides=overrides, debug=config.debug)
+            result = start_baremetal_hosts_with_iso(baremetal_hosts, iso_url, overrides=overrides, debug=config.debug)
+            if result['result'] != 'success':
+                return result
             worker_overrides['workers'] = workers - len(new_baremetal_hosts)
     if platform is not None:
         return {'result': 'success'}
@@ -291,34 +349,25 @@ def create(config, plandir, cluster, overrides):
     elif not os.path.isabs(os.environ['KUBECONFIG']):
         os.environ['KUBECONFIG'] = f"{os.getcwd()}/{os.environ['KUBECONFIG']}"
         if not os.path.exists(os.environ['KUBECONFIG']):
-            msg = "Kubeconfig not found. Leaving..."
-            return {'result': 'failure', 'reason': msg}
+            return {'result': 'failure', 'reason': "Kubeconfig not found. Leaving..."}
     data = safe_load(open(f'{plandir}/kcli_default.yml'))
     data.update(overrides)
     fix_typos(data)
     retries = data.get('retries')
-    if 'cluster' in overrides:
-        clustervalue = overrides.get('cluster')
-    elif cluster is not None:
-        clustervalue = cluster
-    else:
-        clustervalue = 'myhypershift'
-    data['cluster'] = clustervalue
+    data['cluster'] = cluster
     data['kube'] = data['cluster']
     data['kubetype'] = 'hypershift'
-    ignore_hosts = data.get('ignore_hosts', False)
-    pprint(f"Deploying cluster {clustervalue}")
-    plan = cluster if cluster is not None else clustervalue
+    ignore_hosts = data['ignore_hosts'] or data['sslip']
+    pprint(f"Deploying cluster {cluster}")
+    plan = cluster
     upstream = data['upstream']
     platform = data.get('platform')
     assisted = platform == 'assisted'
     data['assisted'] = assisted
-    assisted_vms = data.get('assisted_vms')
     kubevirt = platform == 'kubevirt'
     data['kubevirt'] = kubevirt
     if platform not in [None, 'kubevirt', 'assisted']:
-        msg = "Incorrect platform. Choose between None, kubevirt and assisted"
-        return {'result': 'failure', 'reason': msg}
+        return {'result': 'failure', 'reason': "Incorrect platform. Choose between None, kubevirt and assisted"}
     baremetal_iso = data.get('baremetal_iso', False)
     baremetal_hosts = data.get('baremetal_hosts')
     async_install = data.get('async')
@@ -330,7 +379,13 @@ def create(config, plandir, cluster, overrides):
     data['original_domain'] = domain
     apps = overrides.get('apps', [])
     workers = data.get('workers')
+    use_management_version = False
     version = data.get('version')
+    if version not in ['candidate', 'stable', 'ci', 'nightly', 'latest', 'cluster']:
+        return {'result': 'failure', 'reason': f"Invalid version {version}"}
+    elif version == 'cluster':
+        use_management_version = True
+        version = 'stable'
     coredns = data.get('coredns')
     tag = data.get('tag')
     if str(tag) == '4.1':
@@ -338,8 +393,7 @@ def create(config, plandir, cluster, overrides):
         data['tag'] = tag
     pull_secret = os.path.expanduser(pwd_path(data.get('pull_secret')))
     if not os.path.exists(pull_secret):
-        msg = f"Missing pull secret file {pull_secret}"
-        return {'result': 'failure', 'reason': msg}
+        return {'result': 'failure', 'reason': f"Missing pull secret file {pull_secret}"}
     network_type = data['network_type']
     if network_type not in ['Calico', 'OVNKubernetes']:
         data['network_type'] = 'Other'
@@ -347,8 +401,7 @@ def create(config, plandir, cluster, overrides):
     namespace = data.get('namespace')
     clusterdir = os.path.expanduser(f"~/.kcli/clusters/{cluster}")
     if os.path.exists(clusterdir):
-        msg = f"Remove existing directory {clusterdir} or use --force"
-        return {'result': 'failure', 'reason': msg}
+        return {'result': 'failure', 'reason': f"Remove existing directory {clusterdir} or use --force"}
     else:
         os.makedirs(f"{clusterdir}/auth")
     if which('oc') is None:
@@ -361,14 +414,15 @@ def create(config, plandir, cluster, overrides):
             pprint(f"Using default class {sc['metadata']['name']}")
             default_sc = True
     if not default_sc:
-        msg = "Default Storage class not found. Leaving..."
-        return {'result': 'failure', 'reason': msg}
+        return {'result': 'failure', 'reason': "Default Storage class not found. Leaving..."}
     kubevirt_crd_cmd = 'oc get crd hyperconvergeds.hco.kubevirt.io -o yaml 2>/dev/null'
     if kubevirt and safe_load(os.popen(kubevirt_crd_cmd).read()) is None:
         warning("Kubevirt not fully installed. Installing it for you")
         app_name, source, channel, csv, description, x_namespace, channels, crds = olm_app('kubevirt-hyperconverged')
         app_data = {'name': app_name, 'source': source, 'channel': channel, 'namespace': x_namespace, 'csv': csv}
-        config.create_app_openshift(app_name, app_data)
+        result = config.create_app_openshift(app_name, app_data)
+        if result != 0:
+            return {'result': 'failure', 'reason': "Couldnt install kubevirt properly"}
     kubeconfig = os.environ.get('KUBECONFIG')
     kubeconfigdir = os.path.dirname(kubeconfig) if kubeconfig is not None else os.path.expanduser("~/.kube")
     kubeconfig = os.path.basename(kubeconfig) if kubeconfig is not None else 'config'
@@ -382,20 +436,21 @@ def create(config, plandir, cluster, overrides):
         if mce_hypershift:
             warning("Hypershift needed. Installing it for you")
         if need_assisted:
-            warning("Also installing assisted")
+            warning("Assisted needed. Installing it for you")
         app_name, source, channel, csv, description, x_namespace, channels, crds = olm_app('multicluster-engine')
         app_data = {'name': app_name, 'source': source, 'channel': channel, 'namespace': x_namespace, 'csv': csv,
-                    'mce_hypershift': mce_hypershift, 'assisted': need_assisted, 'version': version, 'tag': tag}
-        config.create_app_openshift(app_name, app_data)
+                    'mce_hypershift': mce_hypershift, 'assisted': need_assisted, 'version': version, 'tag': tag,
+                    'pull_secret': pull_secret}
+        result = config.create_app_openshift(app_name, app_data)
+        if result != 0:
+            return {'result': 'failure', 'reason': "Couldnt install mce properly"}
         sleep(240)
         if assisted and safe_load(os.popen(assisted_crd_cmd).read()) is None:
-            msg = "Couldnt install assisted via mce properly"
-            return {'result': 'failure', 'reason': msg}
+            return {'result': 'failure', 'reason': "Couldnt install assisted via mce properly"}
     if need_hypershift and upstream:
         warning("Hypershift needed. Installing it for you")
         if which('podman') is None:
-            msg = "Please install podman first in order to install hypershift"
-            return {'result': 'failure', 'reason': msg}
+            return {'result': 'failure', 'reason': "Please install podman first in order to install hypershift"}
         else:
             hypercmd = f"podman pull {data['operator_image']}"
             call(hypercmd, shell=True)
@@ -405,8 +460,7 @@ def create(config, plandir, cluster, overrides):
             call(hypercmd, shell=True)
             sleep(120)
     if safe_load(os.popen(hosted_crd_cmd).read()) is None:
-        msg = "Couldnt install hypershift properly"
-        return {'result': 'failure', 'reason': msg}
+        return {'result': 'failure', 'reason': "Couldnt install hypershift properly"}
     registry = 'quay.io'
     management_image = os.popen("oc get clusterversion version -o jsonpath='{.status.desired.image}'").read()
     prefixes = ['quay.io', 'registry.ci']
@@ -436,15 +490,19 @@ def create(config, plandir, cluster, overrides):
     data['registry'] = registry
     data['basedir'] = '/workdir' if container_mode() else '.'
     supported_data = safe_load(os.popen("oc get cm/supported-versions -o yaml -n hypershift").read())
-    if supported_data is not None:
+    if use_management_version:
+        tag = os.popen("oc get clusterversion version -o jsonpath='{.status.desired.version}'").read()
+        pprint(f"Using tag {tag}")
+        if 'ec' in tag:
+            version = 'candidate'
+    elif supported_data is not None:
         supported_data = supported_data['data']
         supported_versions = supported_versions = supported_data['supported-versions']
         versions = safe_load(supported_versions)['versions']
         if version == 'latest':
             tag = versions[0]
         elif str(tag) not in versions:
-            msg = f"Invalid tag {tag}. Choose between {','.join(versions)}"
-            return {'result': 'failure', 'reason': msg}
+            return {'result': 'failure', 'reason': f"Invalid tag {tag}. Choose between {','.join(versions)}"}
     else:
         warning(f"Couldnt verify whether {tag} is a valid tag")
     management_cmd = "oc get ingresscontroller -n openshift-ingress-operator default -o jsonpath='{.status.domain}'"
@@ -488,8 +546,7 @@ def create(config, plandir, cluster, overrides):
     elif os.path.exists(pub_key):
         data['pub_key'] = open(pub_key).read().strip()
     else:
-        msg = f"Publickey file {pub_key} not found"
-        return {'result': 'failure', 'reason': msg}
+        return {'result': 'failure', 'reason': f"Public key file {pub_key} not found"}
     network = data.get('network')
     ingress_ip = data.get('ingress_ip')
     virtual_router_id = None
@@ -501,10 +558,7 @@ def create(config, plandir, cluster, overrides):
             data['domain'] = domain
             pprint(f"Setting domain to {domain}")
             try:
-                cmcmd = "oc patch ingresscontroller -n openshift-ingress-operator default --type=json -p "
-                cmcmd += "'[{ \"op\": \"add\", \"path\": \"/spec/routeAdmission\", "
-                cmcmd += "\"value\": {wildcardPolicy: \"WildcardsAllowed\"}}]'"
-                call(cmcmd, shell=True)
+                patch_ingress_controller_wildcard()
             except:
                 warning("Couldnt patch ingresscontroller to support wildcards. Assuming it's configured properly")
             if not kubevirt:
@@ -519,8 +573,7 @@ def create(config, plandir, cluster, overrides):
                     route_cmd += f"--hostname={hostname} --wildcard-policy=Subdomain --port=443"
                     call(route_cmd, shell=True)
                 elif ingress_ip is None:
-                    msg = f"Couldnt gather an ingress_ip from network {network}"
-                    return {'result': 'failure', 'reason': msg}
+                    return {'result': 'failure', 'reason': f"Couldnt gather an ingress_ip from network {network}"}
         elif provider == 'kvm' and networkinfo['type'] == 'routed':
             cidr = networkinfo['cidr']
             ingress_index = 3 if ':' in cidr else -4
@@ -528,8 +581,7 @@ def create(config, plandir, cluster, overrides):
             warning(f"Using {ingress_ip} as ingress_ip")
             data['ingress_ip'] = ingress_ip
         else:
-            msg = "You need to define ingress_ip in your parameters file"
-            return {'result': 'failure', 'reason': msg}
+            return {'result': 'failure', 'reason': "You need to define ingress_ip in your parameters file"}
     elif kubevirt:
         warning(f"Note that ingress_ip {ingress_ip} won't be configured as you're using kubevirt provider")
     if ingress_ip is not None and data.get('virtual_router_id') is None and not kubevirt_ingress_svc:
@@ -540,7 +592,6 @@ def create(config, plandir, cluster, overrides):
         domain = f"{ingress_ip.replace('.', '-').replace(':', '-')}.sslip.io"
         data['domain'] = domain
         pprint(f"Setting domain to {domain}")
-        ignore_hosts = False
     assetsdata = data.copy()
     copy2(f'{kubeconfigdir}/{kubeconfig}', f"{clusterdir}/kubeconfig.mgmt")
     cidr = data.get('cidr', '192.168.122.0/24')
@@ -568,6 +619,7 @@ def create(config, plandir, cluster, overrides):
                 pprint(f"Injecting manifest {f}")
                 mc_data = json.dumps(mc_data)
                 manifests.append({'name': mc_name, 'data': mc_data})
+    ingress_metallb = data['assisted_ingress_metallb']
     if assisted:
         keepalived_yml_data = config.process_inputfile(cluster, f"{plandir}/staticpods/keepalived.yml", overrides=data)
         keepalived_yml_data = b64encode(keepalived_yml_data.encode()).decode("UTF-8")
@@ -581,17 +633,17 @@ def create(config, plandir, cluster, overrides):
             coredns_conf_data = config.process_inputfile(cluster, f"{plandir}/Corefile", overrides=data)
             coredns_conf_data = b64encode(coredns_conf_data.encode()).decode("UTF-8")
             assisted_data['coredns_conf_data'] = coredns_conf_data
-            force_dns_data = config.process_inputfile(cluster, f"{plandir}/99-forcedns", overrides=data)
+            force_dns_data = config.process_inputfile(cluster, f"{plandir}/99-kcli-forcedns", overrides=data)
             force_dns_data = b64encode(force_dns_data.encode()).decode("UTF-8")
             assisted_data['force_dns_data'] = force_dns_data
         assisted_data['cluster'] = cluster
         assisted_data['sslip'] = sslip
         assisted_data['coredns'] = coredns
-        assisted_data = config.process_inputfile(cluster, f'{plandir}/assisted_ingress.yml', overrides=assisted_data)
-        assisted_data = json.dumps(assisted_data)
-        manifests.append({'name': f'assisted-ingress-{cluster}', 'data': assisted_data})
-    if manifests:
-        assetsdata['manifests'] = manifests
+        if not ingress_metallb:
+            assisted_data = config.process_inputfile(cluster, f'{plandir}/assisted_ingress.yml',
+                                                     overrides=assisted_data)
+            assisted_data = json.dumps(assisted_data)
+            manifests.append({'name': f'assisted-ingress-{cluster}', 'data': assisted_data})
     async_files = []
     async_tempdir = TemporaryDirectory()
     asyncdir = async_tempdir.name
@@ -683,9 +735,12 @@ def create(config, plandir, cluster, overrides):
     if 'OPENSHIFT_INSTALL_RELEASE_IMAGE_OVERRIDE' in os.environ:
         assetsdata['hosted_image'] = os.environ['OPENSHIFT_INSTALL_RELEASE_IMAGE_OVERRIDE']
     else:
-        hosted_version = data['hosted_version'] or version
-        hosted_tag = data['hosted_tag'] or tag
-        assetsdata['hosted_image'] = offline_image(version=hosted_version, tag=hosted_tag, pull_secret=pull_secret)
+        if use_management_version:
+            assetsdata['hosted_image'] = management_image
+        elif version in ['candidate', 'stable', 'ci', 'nightly', 'latest']:
+            assetsdata['hosted_image'] = offline_image(version=version, tag=tag, pull_secret=pull_secret)
+        else:
+            return {'result': 'failure', 'reason': f"Invalid version {version}"}
     hostedclusterfile = config.process_inputfile(cluster, f"{plandir}/hostedcluster.yaml", overrides=assetsdata)
     with open(f"{clusterdir}/hostedcluster.yaml", 'w') as f:
         f.write(hostedclusterfile)
@@ -697,11 +752,10 @@ def create(config, plandir, cluster, overrides):
         if version in ['ci', 'nightly']:
             nightly = version == 'nigthly'
             run = get_ci_installer(pull_secret, tag=tag, nightly=nightly)
-        elif version in ['dev-preview', 'stable', 'latest']:
+        elif version in ['candidate', 'stable', 'latest']:
             run = get_downstream_installer(version=version, tag=tag, pull_secret=pull_secret)
         else:
-            msg = f"Invalid version {version}"
-            return {'result': 'failure', 'reason': msg}
+            return {'result': 'failure', 'reason': f"Invalid version {version}"}
         if run != 0:
             msg = "Couldn't download openshift-install"
             return {'result': 'failure', 'reason': msg}
@@ -711,35 +765,34 @@ def create(config, plandir, cluster, overrides):
     else:
         pprint("Reusing matching openshift-install")
     os.environ["PATH"] = f'{os.getcwd()}:{os.environ["PATH"]}'
-    INSTALLER_VERSION = get_installer_version()
-    pprint(f"Using installer version {INSTALLER_VERSION}")
     if 'OPENSHIFT_INSTALL_RELEASE_IMAGE_OVERRIDE' in os.environ:
-        assetsdata['nodepool_image'] = os.environ['OPENSHIFT_INSTALL_RELEASE_IMAGE_OVERRIDE']
+        nodepool_image = os.environ['OPENSHIFT_INSTALL_RELEASE_IMAGE_OVERRIDE']
+    elif use_management_version:
+        nodepool_image = management_image
     else:
+        INSTALLER_VERSION = get_installer_version()
+        pprint(f"Using installer version {INSTALLER_VERSION}")
         nodepool_image = os.popen("openshift-install version | grep 'release image' | cut -f3 -d' '").read().strip()
-        assetsdata['nodepool_image'] = nodepool_image
     if assisted:
-        if assisted_vms:
-            if data['disk_size'] < 200:
-                data['disk_size'] = 200
-            if data['numcpus'] < 16:
-                data['numcpus'] = 16
-            if data['memory'] < 20480:
-                data['memory'] = 20480
-            worker_threaded = data.get('threaded', False) or data.get('assisted_vms_threaded', False)
+        data['disk_size'] = max(data['disk_size'], 200)
+        data['numcpus'] = max(data['numcpus'], 16)
+        data['memory'] = max(data['memory'], 20480)
+        worker_threaded = data.get('threaded', False)
+        assisted_vms_number = workers - len(baremetal_hosts)
+        if assisted_vms_number > 0:
+            data['assisted_vms_number'] = assisted_vms_number
             result = config.plan(plan, inputfile=f'{plandir}/kcli_plan_assisted.yml', overrides=data,
                                  threaded=worker_threaded)
             if result['result'] != 'success':
                 return result
             vms = result['newvms']
             ksushy_ip = data['assisted_vms_ksushy_ip']
-            ksushy_url = f'http://{ksushy_ip}:9000/redfish/v1/Systems/{config.client}'
+            ksushy_url = f'https://{ksushy_ip}:9000/redfish/v1/Systems/{config.client}'
             virtual_hosts = []
             for vm in vms:
                 new_mac = config.k.info(vm)['nets'][0]['mac']
                 virtual_hosts.append({'url': f'{ksushy_url}/{vm}', 'mac': new_mac})
             baremetal_hosts.extend(virtual_hosts)
-            data['assisted_vms_number'] = len(virtual_hosts)
         create_bmh_objects(config, plandir, cluster, namespace, baremetal_hosts, overrides)
         agents = len(baremetal_hosts) or 2
         pprint(f"Waiting for {agents} agents to appear")
@@ -776,6 +829,11 @@ def create(config, plandir, cluster, overrides):
             if not images:
                 msg = f"Missing {image}. Indicate correct image in your parameters file..."
                 return {'result': 'failure', 'reason': msg}
+    extra_nodepools = data.get('extra_nodepools', [])
+    nodepools = [cluster] + extra_nodepools
+    process_nodepools(config, plandir, namespace, cluster, platform, nodepools, nodepool_image, overrides, manifests)
+    if 'name' in data:
+        del data['name']
     with open(f"{clusterdir}/kcli_parameters.yml", 'w') as p:
         installparam = overrides.copy()
         installparam['client'] = config.client
@@ -797,36 +855,9 @@ def create(config, plandir, cluster, overrides):
         installparam['original_domain'] = data['original_domain']
         if baremetal_hosts:
             installparam['baremetal_hosts'] = baremetal_hosts
-        installparam['assisted_vms_number'] = data['assisted_vms_number']
+        installparam['registry'] = data.get('registry', 'quay.io')
+        installparam['extra_nodepools'] = extra_nodepools
         safe_dump(installparam, p, default_flow_style=False, encoding='utf-8', allow_unicode=True)
-    nodepoolfile = config.process_inputfile(cluster, f"{plandir}/nodepool.yaml", overrides=assetsdata)
-    with open(f"{clusterdir}/nodepool.yaml", 'w') as f:
-        f.write(nodepoolfile)
-    cmcmd = f"oc create -f {clusterdir}/nodepool.yaml"
-    call(cmcmd, shell=True)
-    assetsdata['clusterdir'] = clusterdir
-    if platform is None:
-        ignitionscript = config.process_inputfile(cluster, f"{plandir}/ignition.sh", overrides=assetsdata)
-        with open(f"{clusterdir}/ignition.sh", 'w') as f:
-            f.write(ignitionscript)
-        pprint("Waiting before ignition data is available")
-        user_data = f"user-data-{cluster}"
-        call(f"until oc -n {namespace}-{cluster} get secret | grep {user_data} >/dev/null 2>&1 ; do sleep 1 ; done",
-             shell=True)
-        ignition_worker = f"{clusterdir}/nodepool.ign"
-        timeout = 0
-        while True:
-            if os.path.exists(ignition_worker):
-                break
-            else:
-                sleep(30)
-                timeout += 30
-                if timeout > 300:
-                    msg = "Timeout trying to retrieve worker ignition"
-                    return {'result': 'failure', 'reason': msg}
-            call(f'bash {clusterdir}/ignition.sh', shell=True)
-    if 'name' in data:
-        del data['name']
     autoapproverpath = f'{clusterdir}/autoapprovercron.yml'
     autoapprover = config.process_inputfile(cluster, f"{plandir}/autoapprovercron.yml", overrides=data)
     with open(autoapproverpath, 'w') as f:
@@ -834,18 +865,20 @@ def create(config, plandir, cluster, overrides):
     call(f"oc apply -f {autoapproverpath}", shell=True)
     if platform is None:
         if provider in cloud_providers + ['openstack']:
-            copy2(f"{clusterdir}/nodepool.ign", f"{clusterdir}/nodepool.ign.ori")
+            copy2(f"{clusterdir}/nodepool_{cluster}.ign", f"{clusterdir}/nodepool_{cluster}.ign.ori")
             bucket = f"{cluster}-{domain.replace('.', '-')}"
             if bucket not in config.k.list_buckets():
                 config.k.create_bucket(bucket)
-            config.k.upload_to_bucket(bucket, f"{clusterdir}/nodepool.ign", public=True)
-            bucket_url = config.k.public_bucketfile_url(bucket, "nodepool.ign")
+            config.k.upload_to_bucket(bucket, f"{clusterdir}/nodepool_{cluster}.ign", public=True)
+            bucket_url = config.k.public_bucketfile_url(bucket, "nodepool_{cluster}.ign")
             new_ignition = {'ignition': {'config': {'merge': [{'source': bucket_url}]}, 'version': '3.2.0'}}
-            with open(f"{clusterdir}/nodepool.ign", 'w') as f:
+            with open(f"{clusterdir}/nodepool_{cluster}.ign", 'w') as f:
                 f.write(json.dumps(new_ignition))
         elif baremetal_iso or baremetal_hosts:
             iso_url = handle_baremetal_iso(config, plandir, cluster, data, baremetal_hosts)
-            start_baremetal_hosts(baremetal_hosts, iso_url, overrides=overrides, debug=config.debug)
+            result = start_baremetal_hosts_with_iso(baremetal_hosts, iso_url, overrides=overrides, debug=config.debug)
+            if result['result'] != 'success':
+                return result
             data['workers'] = data.get('workers', 2) - len(baremetal_hosts)
     pprint("Waiting for kubeconfig to be available")
     call(f"until oc -n {namespace} get secret {cluster}-admin-kubeconfig >/dev/null 2>&1 ; do sleep 1 ; done",
@@ -887,7 +920,7 @@ def create(config, plandir, cluster, overrides):
     async_tempdir.cleanup()
     if ignore_hosts:
         warning("Not updating /etc/hosts as per your request")
-    else:
+    elif ingress_ip is not None:
         update_hypershift_etc_hosts(cluster, domain, ingress_ip)
     if provider in cloud_providers:
         result = config.plan(plan, inputfile=f'{plandir}/cloud_lb_apps.yml', overrides=data)
@@ -913,6 +946,16 @@ def create(config, plandir, cluster, overrides):
     apps = overrides.get('apps', [])
     overrides['hypershift'] = True
     overrides['cluster'] = cluster
+    if ingress_metallb:
+        if 'metallb-operator' not in apps:
+            apps.append('metallb-operator')
+        if 'metallb_ranges' not in overrides:
+            overrides['metallb_ranges'] = [ingress_ip]
+        ingress_file = config.process_inputfile('xxx', f'{plandir}/assisted_ingress_service.yml.j2',
+                                                overrides={'ingress_ip': ingress_ip})
+        with open(f"{clusterdir}/assisted_ingress_service.yml", 'w') as f:
+            f.write(ingress_file)
+        call(f'oc create -f {clusterdir}/assisted_ingress_service.yml', shell=True)
     process_apps(config, clusterdir, apps, overrides)
     if autoscale:
         config.import_in_kube(network=network, secure=True)
